using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Brume;

namespace BrumeParticles.Particles
{
    /// <summary>
    /// 
    /// </summary>
    public class ParticleSystem
    {
        // Fields
        private List<int> ActiveParticles;
        public Color Color;
        public Color ColorNoise;
        public IParticleEmitter Emitter;

        public float LifeTime;
        public float LifeTimeNoise;
        public int MaxParticles;
        public int MinParticles;
        private List<Particle> ParticlesStore;
        public int ParticlesToAdd;
        public List<IParticleProcessor> Processors;
        private Queue<int> QueuedParticles;
        public float Scale;
        public float ScaleNoise;
        private float TimeElapsedAfterLastAdd;
        BrumeTexture _texture;
        BrumeLightedObject _obj;

        Brume.Brume _brume;

        BrumeObject _attachedObject;

        /// <summary>
        /// Gets the <see cref="BrumeParticles.Particles.Particle"/> at the specified index.
        /// </summary>
        /// <value></value>
        public Particle this[int Index]
        {
            get
            {
                return this.ParticlesStore[this.ActiveParticles[Index]];
            }
        }

        /// <summary>
        /// Gets the particles count.
        /// </summary>
        /// <value>The particles count.</value>
        public int ParticlesCount
        {
            get
            {
                return this.ActiveParticles.Count;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ParticleSystem"/> class.
        /// </summary>
        /// <param name="brume">The brume.</param>
        /// <param name="Texture">The texture.</param>
        /// <param name="MinParticles">The min particles.</param>
        /// <param name="MaxParticles">The max particles.</param>
        /// <param name="LifeTime">The life time.</param>
        /// <param name="LifeTimeNoise">The life time noise.</param>
        /// <param name="Color">The color.</param>
        /// <param name="Emitter">The emitter.</param>
        public ParticleSystem(Brume.Brume brume, BrumeTexture Texture, int MinParticles, int MaxParticles, float LifeTime, float LifeTimeNoise, Color Color, IParticleEmitter Emitter)
            : this(brume, null, Texture, MinParticles, MaxParticles, LifeTime, LifeTimeNoise, Color, Emitter)
        {

        }


        /// <summary>
        /// Initializes a new instance of the <see cref="ParticleSystem"/> class.
        /// </summary>
        /// <param name="brume">The brume.</param>
        /// <param name="attachedObject">The attached object.</param>
        /// <param name="Texture">The texture.</param>
        /// <param name="MinParticles">The min particles.</param>
        /// <param name="MaxParticles">The max particles.</param>
        /// <param name="LifeTime">The life time.</param>
        /// <param name="LifeTimeNoise">The life time noise.</param>
        /// <param name="Color">The color.</param>
        /// <param name="Emitter">The emitter.</param>
        public ParticleSystem(Brume.Brume brume, BrumeObject attachedObject, BrumeTexture Texture, int MinParticles, int MaxParticles, float LifeTime, float LifeTimeNoise, Color Color, IParticleEmitter Emitter)
        {
            _brume = brume;
            _attachedObject = attachedObject;

            _obj = new BrumeLightedObject(brume, "particles" + Guid.NewGuid().ToString());

            _obj.Mesh = new BrumeSquareMeshTextured(brume, "particles" + Guid.NewGuid().ToString(), 0, 0);
            _obj.Mesh.Init();
            _obj.RenderEvent += new BrumeObjectSimpleDelegate(obj_RenderEvent);
            _obj.MoveEvent += new BrumeObjectMoveDelegate(_obj_MoveEvent);
            _obj.Size = 0.0000000001f;

            _obj.RenderStates.Texture[0] = Texture;
            _obj.RenderStates.AlphaBlending = BrumeRenderParams.ALPHABLENDING_TEXTUREALPHA;
            _obj.RenderStates.UseOwnAmbientLight = true;


            _texture = Texture;
            this.MinParticles = MinParticles;
            this.MaxParticles = MaxParticles;
            this.Emitter = Emitter;
            this.LifeTime = LifeTime;
            this.LifeTimeNoise = LifeTimeNoise;
            this.Color = Color;
            this.ColorNoise = Color.Black;
            this.Scale = 1f;
            this.ScaleNoise = 0f;
            this.Processors = new List<IParticleProcessor>();
            this.ParticlesStore = new List<Particle>(MaxParticles);
            this.ActiveParticles = new List<int>(MaxParticles);
            this.QueuedParticles = new Queue<int>(MaxParticles);
            this.TimeElapsedAfterLastAdd = 0f;
        }

        /// <summary>
        /// _obj_s the move event.
        /// </summary>
        /// <param name="obj">The obj.</param>
        /// <param name="fElapsedTime">The f elapsed time.</param>
        void _obj_MoveEvent(BrumeObject obj, float fElapsedTime)
        {
            Update(fElapsedTime);
        }

        /// <summary>
        /// Adds the particle.
        /// </summary>
        public void AddParticle()
        {
            this.ParticlesToAdd++;
        }

        /// <summary>
        /// Adds the processor.
        /// </summary>
        /// <param name="Processor">The processor.</param>
        public void AddProcessor(IParticleProcessor Processor)
        {
            this.Processors.Add(Processor);
        }



        /// <summary>
        /// Generates the particle.
        /// </summary>
        private void GenerateParticle()
        {
            Particle item = new Particle(this.Emitter.GeneratePosition(),
                                        this.Emitter.GenerateDirection(),
                                        Randomizer.GetRandomColor(RemoveColor(this.Color, this.ColorNoise),
                                                                   AddColor(this.Color, this.ColorNoise)),
                                        Randomizer.GetRandomFloat(this.Scale - this.ScaleNoise, this.Scale + this.ScaleNoise), 1f, Randomizer.GetRandomFloat(this.LifeTime - this.LifeTimeNoise, this.LifeTime + this.LifeTimeNoise), this.Emitter.GetSpeed());
            for (int i = 0; i < this.Processors.Count; i++)
            {
                item.ProcessorsState.AddElement(eParticleProcessorState.Dead);
            }
            this.ParticlesStore.Add(item);
        }



        /// <summary>
        /// Obj_s the render event.
        /// </summary>
        /// <param name="obj">The obj.</param>
        void obj_RenderEvent(BrumeObject obj)
        {

            BrumeMatrix world = BrumeMatrix.Identity;

            BrumeMatrix newWorld = world;

            // no parent size
            newWorld.M11 *= 1.0f / newWorld.M11;
            newWorld.M22 *= 1.0f / newWorld.M22;
            newWorld.M33 *= 1.0f / newWorld.M33;


            // bilbording
            BrumeMatrix local = _brume.ActiveCamera.BillboardMatrix;

            local.M41 = 0.0f;
            local.M42 = 0.0f;
            local.M43 = 0.0f;
            local.M44 = 1.0f;

            BrumeMatrix goodLocal = local;



            for (int i = 0; i < this.ParticlesStore.Count; i++)
            {
                _brume.GraphicApi.SetAmbientLight(ParticlesStore[i].Color);
                local = goodLocal;

                local = BrumeMatrix.Scaling(ParticlesStore[i].Scale, ParticlesStore[i].Scale, ParticlesStore[i].Scale) * local;

                local.M41 = ParticlesStore[i].Position.X;
                local.M42 = ParticlesStore[i].Position.Y;
                local.M43 = ParticlesStore[i].Position.Z;

                _brume.GraphicApi.SetWorld(local * newWorld);
                obj.Mesh.Render();
            }

        }

        /// <summary>
        /// Resets the particle.
        /// </summary>
        /// <param name="Index">The index.</param>
        private void ResetParticle(int Index)
        {
            this.ParticlesStore[Index].Reset(this.Emitter.GeneratePosition(), this.Emitter.GenerateDirection(), Randomizer.GetRandomColor(RemoveColor(this.Color, this.ColorNoise), AddColor(this.Color, this.ColorNoise)), Randomizer.GetRandomFloat(this.Scale - this.ScaleNoise, this.Scale + this.ScaleNoise), 1f, Randomizer.GetRandomFloat(this.LifeTime - this.LifeTimeNoise, this.LifeTime + this.LifeTimeNoise), this.Emitter.GetSpeed());
            this.ParticlesStore[Index].ResetProcessors();
        }





        /// <summary>
        /// Updates the specified elapsed time.
        /// </summary>
        /// <param name="elapsedTime">The elapsed time.</param>
        public void Update(float elapsedTime)
        {
            if (_attachedObject != null)
            {
                Emitter.Center = _attachedObject.Pos;
            }

            if (this.ParticlesStore.Count == 0)
            {
                for (int i = 0; i < this.MaxParticles; i++)
                {
                    this.GenerateParticle();
                    this.QueuedParticles.Enqueue(i);
                }
            }
            float deltaTime = elapsedTime;
            this.TimeElapsedAfterLastAdd += deltaTime;
            if (this.ParticlesToAdd > this.QueuedParticles.Count)
            {
                this.ParticlesToAdd = this.QueuedParticles.Count;
            }
            if (this.ParticlesToAdd > 0)
            {
                if ((this.ParticlesToAdd + this.ActiveParticles.Count) >= this.MaxParticles)
                {
                    this.ParticlesToAdd = this.MaxParticles - this.ActiveParticles.Count;
                }
                if (this.Emitter.Mode != eParticleEmissionMode.Manual)
                {
                    int item = this.QueuedParticles.Dequeue();
                    this.ActiveParticles.Add(item);
                    this.TimeElapsedAfterLastAdd = 0f;
                    this.ParticlesToAdd--;
                }
                else
                {
                    for (int j = 0; j < this.ParticlesToAdd; j++)
                    {
                        int num5 = this.QueuedParticles.Dequeue();
                        this.ActiveParticles.Add(num5);
                        this.TimeElapsedAfterLastAdd = 0f;
                    }
                }
            }
            else if (this.Emitter.Mode != eParticleEmissionMode.Manual)
            {
                if (this.ActiveParticles.Count < this.MinParticles)
                {
                    this.ParticlesToAdd = this.MinParticles - this.ActiveParticles.Count;
                }
                else if ((this.ActiveParticles.Count < this.MaxParticles) && (this.TimeElapsedAfterLastAdd >= this.Emitter.EmitterRate))
                {
                    this.ParticlesToAdd = (int)(this.TimeElapsedAfterLastAdd / this.Emitter.EmitterRate);
                }
            }
            if (this.ActiveParticles.Count > 0)
            {
                for (int k = 0; k < this.ActiveParticles.Count; k++)
                {
                    if (this.ParticlesStore[this.ActiveParticles[k]].Update(deltaTime))
                    {
                        for (int m = 0; m < this.Processors.Count; m++)
                        {
                            if (this.Processors[m].CheckState(this.ParticlesStore[this.ActiveParticles[k]].NormalizedLifeTime, this.ParticlesStore[this.ActiveParticles[k]].ProcessorsState[m]) == eParticleProcessorState.JustAlive)
                            {
                                this.Processors[m].Initialize(this.ParticlesStore[this.ActiveParticles[k]]);
                            }
                            this.ParticlesStore[this.ActiveParticles[k]].ProcessorsState.Set(m, this.Processors[m].State);
                            if (this.Processors[m].State == eParticleProcessorState.Alive)
                            {
                                this.Processors[m].Process(this.ParticlesStore[this.ActiveParticles[k]], deltaTime);
                            }
                        }
                    }
                    else
                    {
                        if (this.Emitter.Mode == eParticleEmissionMode.Continious)
                        {
                            this.ResetParticle(this.ActiveParticles[k]);
                        }

                        this.QueuedParticles.Enqueue(this.ActiveParticles[k]);
                        this.ActiveParticles.RemoveAt(k);

                        if (ActiveParticles.Count == 0 && Emitter.Mode == eParticleEmissionMode.Once)
                            Remove();
                    }
                }
            }
        }


        /// <summary>
        /// Removes this instance.
        /// </summary>
        public void Remove()
        {
            _brume.RemoveObject(_obj);

        }




        #region Colors Utils

        private static Color RemoveColor(Color color1, Color color2)
        {
            return Color.FromArgb(color1.R - color2.R, color1.G - color2.G, color1.B - color2.B);

        }

        private static Color AddColor(Color color1, Color color2)
        {
            return Color.FromArgb(color1.R + color2.R, color1.G + color2.G, color1.B + color2.B);

        }


        #endregion


    }


}
