float4x4 worldViewProj  : WorldViewProjection;
float4x4 worldView      : WorldView;

float3 lightPos;
float3 cameraPos;
float shininess; 

texture diffuseTex;
texture specularTex;
texture edgeTex;

float4 matDiffuse;
float4 matSpecular;


sampler DiffuseSampler = sampler_state
{
    Texture   = <diffuseTex>;
    MipFilter = POINT;
    MinFilter = POINT;
    MagFilter = POINT;
};

sampler SpecularSampler = sampler_state
{
    Texture   = <specularTex>;
    MipFilter = POINT;
    MinFilter = POINT;
    MagFilter = POINT;
};

sampler EdgeSampler = sampler_state
{
    Texture   = <edgeTex>;
    MipFilter = POINT;
    MinFilter = POINT;
    MagFilter = POINT;
};


struct VS_Input
{
   float3 vertexPos  : POSITION;
   float3 normal     : NORMAL;
};


struct Vs_Output
{
   float4 vertexPos  : POSITION;
   float3 color      : COLOR0;
   float2 diffuse     : TEXCOORD0;
   float2 specular    : TEXCOORD1;
   float2 edge        : TEXCOORD2;
};


struct Ps_Output
{
	float4 color      : COLOR;
};


Vs_Output VertexShaderEffect(VS_Input IN)
{
   Vs_Output vs_out;

   // Set output vertex pos, and texture coords.
   vs_out.vertexPos = mul(float4(IN.vertexPos, 1), worldViewProj);
   vs_out.color = float3(0.8f, 0.9f, 1.0f);
   
   // Calculate light vector.
   float3 pos = mul(IN.vertexPos, worldView);
   float3 lightVec = normalize(lightPos - pos);

   // Renormalize the normal.
   float3 n = mul(IN.normal, (float3x3)worldView);
   
   // Calculate the diffuse term (n dot l).
   vs_out.diffuse = max(dot(n, lightVec), 0);

   // Get the eye vector and half angle.
   float3 eyeVec = normalize(cameraPos - pos);
   float3 halfAngle = normalize(lightVec + eyeVec);
   
   // Calculate specular term (pow(max(n dot h), 0) shinines).
   vs_out.specular = pow(max(dot(n, halfAngle), 0), shininess);

   // if no diffuse, then no specular.
   if(vs_out.diffuse.x <= 0) vs_out.specular.x = 0;

   // Get the edge value.
   vs_out.edge = max(dot(n, eyeVec), 0);
   
   return vs_out;
}


Ps_Output PixelShaderEffect(Vs_Output IN)
{
   Ps_Output ps_out;
   
   float diffuseCol = tex1D(DiffuseSampler, IN.diffuse).x;
   float specularCol = tex1D(SpecularSampler, IN.specular).x;
   float edgeCol = tex1D(EdgeSampler, IN.edge).x;
   
   ps_out.color = float4(IN.color * (edgeCol * (matDiffuse * diffuseCol + matSpecular * specularCol)), 1);

   return ps_out;
}


technique CartoonShader
{
   pass Pass0
   {
      VertexShader = compile vs_2_0 VertexShaderEffect();
      PixelShader  = compile ps_2_0 PixelShaderEffect();
   }
}
